@@ -5,11 +5,12 @@ import logging
 import os
 from pathlib import Path
 import subprocess
-from subprocess import CalledProcessError, TimeoutExpired
+from subprocess import CalledProcessError
 import sys
-from typing import IO, Optional, Sequence, Tuple, Union, cast
+import tempfile
+from typing import Generic, IO, Optional, Sequence, Tuple, Type, TypeVar, Union, cast
 
-from subby.utils import create_and_open_tempfile, command_lists_to_strings
+from subby.utils import command_lists_to_strings
 
 LOG = logging.getLogger()
 DEFAULT_EXECUTABLE = "/bin/bash"
@@ -19,6 +20,7 @@ class StdType(enum.IntEnum):
     """
     stdout/stderr types
     """
+
     PIPE = 0
     FILE = 1
     BUFFER = 2
@@ -26,7 +28,10 @@ class StdType(enum.IntEnum):
     OTHER = 4
 
 
-class Processes:
+Mode = TypeVar("Mode", str, bytes)
+
+
+class Processes(Generic[Mode]):
     """
     Encapsulates one or more commands, runs those commands using the
     `subprocess` module, and provides access to the results.
@@ -48,23 +53,51 @@ class Processes:
             `stdout` for details).
         capture_stderr: Whether to capture the contents of stderr of processes
             other than the final process.
+        mode: Mode to use for stdin/stdout/stderr; can be bytes (raw mode) or str
+            (text mode).
+        encoding: Encoding to use when `mode = str`.
         echo: Whether to echo commands to the logger before running the commands.
             Can be overridden by the `run()` method's `echo` parameter.
+        allowed_return_codes: Sequence of return codes that signal successful
+            completion; [0] by default.
+        raise_on_error: Whether to raise an exception if the process terminates
+            with a non-allowed return code.
+        timeout: When blocking, number of seconds to wait for the process to complete
+            before raising a TimeoutError (can be overridden by `timeout` parameter
+            to `block()`).
         popen_kwargs: Keyword arguments to pass to Popen constructors.
-
-    Todo:
-        Add ability to send input to stdin of first process.
     """
+
     def __init__(
         self,
         cmds: Sequence[Union[str, Sequence[str]]],
-        stdin: Optional[Union[bytes, Path, StdType]] = None,
+        stdin: Optional[Union[Mode, Path, StdType]] = None,
         stdout: Optional[Union[Path, StdType]] = StdType.PIPE,
         stderr: Optional[Union[Path, StdType]] = StdType.PIPE,
         capture_stderr: bool = True,
+        mode: Type[Mode] = str,
+        encoding: str = "UTF-8",
         echo: bool = None,
-        **popen_kwargs
+        allowed_return_codes: Sequence[int] = (0,),
+        raise_on_error: Optional[bool] = None,
+        timeout: Optional[int] = None,
+        **popen_kwargs,
     ):
+        if "universal_newlines" in popen_kwargs:
+            universal_newlines = popen_kwargs.pop("universal_newlines")
+            if universal_newlines != (mode is str):
+                raise ValueError(
+                    f"'universal_newlines' is redundant; if specified, its value must "
+                    f"match the 'mode' value of {mode}"
+                )
+        if "text" in popen_kwargs:
+            text = popen_kwargs.pop("text")
+            if text != (mode is str):
+                raise ValueError(
+                    f"'text' is redundant; if specified, its value must "
+                    f"match the 'mode' value of {mode}"
+                )
+
         self.cmds = cmds
         self._stdin_arg = stdin
         self._stdin = None
@@ -77,10 +110,15 @@ class Processes:
         self._stderr_type = None
         self.capture_stderr = capture_stderr
         self._stderr_buffers = [] if capture_stderr else None
-        self._stderr_bytes = None
+        self._stderr_content = None
+        self._mode = mode
+        self.encoding = encoding
         self._closed = False
         self.echo = echo
+        self.allowed_return_codes = set(allowed_return_codes)
         self.popen_kwargs = popen_kwargs
+        self.raise_on_error = raise_on_error
+        self.timeout = timeout
         self._processes = None
         self._output_handle = None
         self._returncode = None
@@ -101,7 +139,10 @@ class Processes:
             # Note that if any process other than the last has a return code of
             # None, we ignore it.
             self._returncode = self._processes[-1].poll()
-            if self._returncode == 0 and len(self._processes) > 1:
+            if (
+                self._returncode in self.allowed_return_codes
+                and len(self._processes) > 1
+            ):
                 # The last process finished running without error, but check all
                 # the other processes for an error.
                 for proc in reversed(self._processes[-2::-1]):
@@ -111,35 +152,33 @@ class Processes:
                         break
         return self._returncode
 
+    @property
+    def text_mode(self) -> bool:
+        return self._mode is str
+
     def _init_stdin(self) -> Union[int, IO]:
-        self._stdin, self._stdin_type, retval = Processes._init_std(
-            self._stdin_arg,
-            sys.stdin,
-            False
+        self._stdin, self._stdin_type, retval = self._init_std(
+            self._stdin_arg, sys.stdin, False
         )
         return retval
 
     def _init_stdout(self) -> Union[int, IO]:
-        self._stdout, self._stdout_type, retval = Processes._init_std(
-            self._stdout_arg,
-            sys.stdout,
-            True
+        self._stdout, self._stdout_type, retval = self._init_std(
+            self._stdout_arg, sys.stdout, True
         )
         return retval
 
     def _init_stderr(self) -> Union[int, IO]:
-        self._stderr, self._stderr_type, retval = Processes._init_std(
-            self._stderr_arg,
-            sys.stderr,
-            True
+        self._stderr, self._stderr_type, retval = self._init_std(
+            self._stderr_arg, sys.stderr, True
         )
         return retval
 
-    @staticmethod
     def _init_std(
-        value: Optional[Union[Path, StdType, bytes]],
+        self,
+        value: Optional[Union[Path, StdType, bytes, str]],
         sys_stream: IO,
-        is_output: bool = True
+        is_output: bool = True,
     ) -> Tuple[Optional[IO], StdType, Union[int, IO]]:
         """
 
@@ -148,23 +187,35 @@ class Processes:
         stream = None
         retval = None
         if isinstance(value, bytes):
-            # Put input in a tempfile
             std_type = StdType.BUFFER
-            stream = create_and_open_tempfile("w+b")
-            stream.write(cast(bytes, value))
+            # Put input in a tempfile
+            stream = self._create_and_open_tempfile("w+")
+            if self.text_mode:
+                stream.write(cast(bytes, value).decode(self.encoding))
+            else:
+                stream.write(cast(bytes, value))
+            stream.seek(0)
+        elif isinstance(value, str):
+            std_type = StdType.BUFFER
+            # Put input in a tempfile
+            stream = self._create_and_open_tempfile("w+")
+            if self.text_mode:
+                stream.write(cast(str, value))
+            else:
+                stream.write(cast(str, value).encode(self.encoding))
             stream.seek(0)
         elif isinstance(value, StdType):
             std_type = cast(StdType, value)
             if std_type is StdType.PIPE:
                 retval = subprocess.PIPE
             elif std_type is StdType.BUFFER and is_output:
-                stream = create_and_open_tempfile()
+                stream = self._create_and_open_tempfile()
             elif std_type is StdType.SYS:
                 stream = sys_stream
             else:
                 raise ValueError(f"Invalid argument: {value}")
         elif value is not None:
-            stream = open(value, "wb" if is_output else "rb")
+            stream = self._open_file(cast(Path, value), "w" if is_output else "r")
             std_type = StdType.FILE
 
         return stream, std_type, retval or stream
@@ -177,18 +228,29 @@ class Processes:
             The opened file object.
         """
         if self.capture_stderr:
-            handle = create_and_open_tempfile()
+            handle = self._create_and_open_tempfile()
             self._stderr_buffers.append(handle)
             return handle
 
+    def _open_file(self, path: Union[str, Path], mode: str) -> IO:
+        mode += "t" if self.text_mode else "b"
+        return open(path, mode)
+
+    def _create_and_open_tempfile(self, mode: str = "w") -> IO:
+        """
+        Creates and returns a temporary output file object.
+        """
+        mode += "t" if self.text_mode else "b"
+        return open(tempfile.mkstemp()[1], mode)
+
     @property
-    def output(self) -> bytes:
+    def output(self) -> Mode:
         """
         The contents of the `stdout` stream of the last process in the chain.
         Only available if `self.closed is True` and `self._stdout_type in
         {StdType.PIPE, StdType.BUFFER}`.
         """
-        if not (self.closed and self._stdout_type in {StdType.PIPE, StdType.BUFFER}):
+        if not (self.closed and self.stdout_type in {StdType.PIPE, StdType.BUFFER}):
             raise RuntimeError(
                 "'output' is only available for closed processes using a pipe or "
                 "buffer for stdout."
@@ -196,19 +258,45 @@ class Processes:
         return self._out
 
     @property
-    def error(self) -> bytes:
+    def error(self) -> Mode:
         """
         The contents of the `stderr` stream of the last process in the chain.
         Only available if `self.closed is True` and `self._stderr_type in
         {StdType.PIPE, StdType.BUFFER}`.
         """
-        if not (self.closed and self._stderr_type in {StdType.PIPE, StdType.BUFFER}):
+        if not (self.closed and self.stderr_type in {StdType.PIPE, StdType.BUFFER}):
             raise RuntimeError(
                 "'error' is only available for closed processes using a pipe or "
                 "buffer for stderr."
             )
         return self._err
 
+    @property
+    def stdin_type(self) -> StdType:
+        if self._stdin_type is None:
+            raise RuntimeError("Cannot access 'stdin_type' until after calling 'run'.")
+        return self._stdin_type
+
+    @property
+    def stdin_stream(self) -> IO:
+        """
+        The `stdin` stream.
+
+        Note:
+            Writing directly to the process stdin could cause a deadlock. Unless you
+            know what you are doing, do not access this stream - instead, pass
+            `stdin=<string or bytes>` to the initializer.
+        """
+        if not self.was_run:
+            raise RuntimeError("Cannot access 'stdout' until after calling 'run'.")
+        return self._stdin or self._processes[0].stdin
+
+    @property
+    def stdout_type(self) -> StdType:
+        if self._stdout_type is None:
+            raise RuntimeError("Cannot access 'stdout_type' until after calling 'run'.")
+        return self._stdout_type
+
     @property
     def stdout_stream(self) -> IO:
         """
@@ -218,6 +306,12 @@ class Processes:
             raise RuntimeError("Cannot access 'stdout' until after calling 'run'.")
         return self._stdout or self._processes[-1].stdout
 
+    @property
+    def stderr_type(self) -> StdType:
+        if self._stdout_type is None:
+            raise RuntimeError("Cannot access 'stderr_type' until after calling 'run'.")
+        return self._stderr_type
+
     @property
     def stderr_stream(self) -> IO:
         """
@@ -227,7 +321,7 @@ class Processes:
             raise RuntimeError("Cannot access 'stderr' until after calling 'run'.")
         return self._stderr or self._processes[-1].stderr
 
-    def get_all_stderr(self) -> Sequence[bytes]:
+    def get_all_stderr(self) -> Sequence[Mode]:
         """
         Get the contents of all stderr streams. Stderr streams of all but the final
         process are only available if `self.capture_stderr is True`. Stderr stream
@@ -245,7 +339,7 @@ class Processes:
             )
         stderr = []
         if self.capture_stderr:
-            stderr.extend(self._stderr_bytes)
+            stderr.extend(self._stderr_content)
         if self._stderr_type in (StdType.BUFFER, StdType.PIPE):
             stderr.append(self.error)
         return stderr
@@ -267,9 +361,10 @@ class Processes:
     @property
     def ok(self) -> bool:
         """
-        Whether the commands completed successfully, i.e. all had `returncode == 0`.
+        Whether the commands completed successfully, i.e. all had an allowed
+        returncode.
         """
-        return self.returncode == 0
+        return self.returncode in self.allowed_return_codes
 
     @property
     def closed(self) -> bool:
@@ -302,6 +397,9 @@ class Processes:
         for i, cmd in enumerate(self.cmds, 1):
             popen_kwargs = copy.copy(self.popen_kwargs)
             popen_kwargs.update(kwargs)
+            if self.text_mode:
+                popen_kwargs["universal_newlines"] = True
+                popen_kwargs["encoding"] = self.encoding
             if i == 1:
                 popen_kwargs["stdin"] = self._init_stdin()
             elif procs:
@@ -322,8 +420,8 @@ class Processes:
     def block(
         self,
         close: bool = True,
-        raise_on_error: bool = True,
-        timeout: Optional[int] = None
+        raise_on_error: Optional[bool] = None,
+        timeout: Optional[int] = None,
     ):
         """
         Wait for all commands to finish.
@@ -340,25 +438,29 @@ class Processes:
         if self.closed:
             raise RuntimeError("Cannot call block() after calling close()")
 
+        default_value = "" if self.text_mode else b""
         last_proc = self._processes[-1]
+        if timeout is None:
+            timeout = self.timeout
+
         try:  # TODO: figure out how to test this
             out, err = (
-                b"" if std is None else std.strip()
+                default_value if std is None else std.strip()
                 for std in last_proc.communicate(timeout=timeout)
             )
             if self._stdout_type == StdType.PIPE:
                 self._out = out
             if self._stderr_type == StdType.PIPE:
                 self._err = err
-        except ValueError:
+        except ValueError:  # TODO: figure out how to test this
             LOG.exception("Error reading from stdout/stderr")
-        except TimeoutExpired:
-            LOG.exception(f"Process timed out after {timeout} seconds")
 
         if close and not self.closed:
             self._close_and_set_std()
 
-        if raise_on_error:
+        if raise_on_error is None:
+            raise_on_error = self.raise_on_error
+        if raise_on_error is not False:
             self.raise_if_error()
 
     def kill(self) -> bool:
@@ -414,6 +516,7 @@ class Processes:
         Close any open files and set the values of `self._out` and `self._err`
         if they are of type StdType.BUFFER.
         """
+
         def close_file(handle):
             try:  # TODO: figure out how to test
                 handle.close()
@@ -433,10 +536,10 @@ class Processes:
                 if self._stdin_type is StdType.BUFFER:
                     remove_file(self._stdin)
 
-        def close_output_buffer(handle) -> bytes:
+        def close_output_buffer(handle) -> Mode:
             close_file(handle)
             try:
-                with open(handle.name, "rb") as inp:
+                with self._open_file(handle.name, "r") as inp:
                     return inp.read()
             finally:
                 remove_file(handle)
@@ -452,7 +555,7 @@ class Processes:
             self._err = close_output_buffer(self._stderr)
 
         if self.capture_stderr:
-            self._stderr_bytes = [
+            self._stderr_content = [
                 close_output_buffer(buf) for buf in self._stderr_buffers
             ]
             self._stderr_buffers = None
@@ -468,12 +571,11 @@ class Processes:
             CalledProcessError
         """
         if self.done and not self.ok:
+            sep = "\n" if self.text_mode else b"\n"
             msg = "stderr from executed commands:\n{}".format(
-                b"\n".join(self.get_all_stderr())
-            )
-            raise CalledProcessError(
-                self.returncode, str(self), output=msg
+                sep.join(self.get_all_stderr())
             )
+            raise CalledProcessError(self.returncode, str(self), output=msg)
 
     def __str__(self) -> str:
         cmd_str = " | ".join(command_lists_to_strings(self.cmds))
