@@ -1,13 +1,5 @@
 import shlex
-import tempfile
-from typing import IO, Sequence, Union
-
-
-def create_and_open_tempfile(mode: str = "wb") -> IO:
-    """
-    Creates and returns a temporary output file object.
-    """
-    return open(tempfile.mkstemp()[1], mode)
+from typing import Sequence, Union
 
 
 def quote_args(seq: Sequence[str]) -> str:
@@ -35,10 +27,7 @@ def command_strings_to_lists(
     Returns:
         A sequence of command argument sequences.
     """
-    return [
-        shlex.split(cmd)if isinstance(cmd, str) else cmd
-        for cmd in cmds
-    ]
+    return [shlex.split(cmd) if isinstance(cmd, str) else cmd for cmd in cmds]
 
 
 def command_lists_to_strings(
@@ -53,7 +42,4 @@ def command_lists_to_strings(
     Returns:
         A sequence of command strings.
     """
-    return [
-        quote_args(cmd) if not isinstance(cmd, str) else cmd
-        for cmd in cmds
-    ]
+    return [quote_args(cmd) if not isinstance(cmd, str) else cmd for cmd in cmds]
